//! Sled-style error handling is based on Result<Result<Value, LocalError>, FatalError>.
//! FatalErrors do not get resolved. LocalErrors get bubbled up until they can be handled.

use super::*;
use std::collections::HashMap;

pub struct DB(HashMap<String, i8>);

impl DB {
    pub fn new(lst: Vec<(String, i8)>) -> Self {
        Self(lst.into_iter().collect::<HashMap<String, i8>>())
    }

    /// Retrieve a value from the database. Throw a fatal error with the "fail" key, but
    /// otherwise return either the value or DatabaseError::NotFound.
    ///
    /// ```rust
    /// use error_training::{*, sled::*};
    ///
    /// let db = DB::new(vec![("a".to_owned(), 15), ("b".to_owned(), 0)]);
    /// assert_eq!(db.get("fail"), Err(FatalError::DatabaseCorruption));
    /// assert_eq!(db.get("a"), Ok(Ok(15)));
    /// assert_eq!(db.get("c"), Ok(Err(DatabaseError::NotFound)));
    /// ```
    pub fn get(&self, key: &str) -> Result<Result<i8, DatabaseError>, FatalError> {
        if key == "fail" {
            Err(FatalError::DatabaseCorruption)
        } else {
            Ok(self.0.get(key).copied().ok_or(DatabaseError::NotFound))
        }
    }
}

fn op(val: i8) -> Result<i8, MathError> {
    if val as i32 + 120_i32 > (i8::MAX as i32) {
        Err(MathError::ExceedsMaxint)
    } else {
        Ok(val + 120)
    }
}

/// This function exists to test several of the major cases. This is where I will figure out
/// how to handle everything and also have clean code.
/// ```rust
/// use error_training::{*, sled::*};
///
/// let db = DB::new(vec![("a".to_owned(), 15), ("b".to_owned(), 0)]);
/// assert_eq!(run_op(&db, "a"), Ok(Ok(i8::MAX)));
/// assert_eq!(run_op(&db, "b"), Ok(Ok(120)));
/// assert_eq!(run_op(&db, "c"), Ok(Ok(0)));
/// assert_eq!(run_op(&db, "fail"), Err(FatalError::DatabaseCorruption));
/// ```
///
/// I have defined this function such that a database miss becomes a 0 and no operation will be
/// performed on it. Since that is the only database error that can occur, this function can
/// only return a `MathError` or a `FatalError`.
pub fn run_op(db: &DB, key: &str) -> Result<Result<i8, MathError>, FatalError> {
    let res = match db.get(key)? {
        Err(DatabaseError::NotFound) => Ok(0),
        Ok(val) => op(val),
    };

    Ok(res.or_else(|err| match err {
        MathError::ExceedsMaxint => Ok(127),
        err => Err(err),
    }))
}
